@@ -33,15 +33,20 @@ struct connman_device {
 	enum connman_device_type type;
 	enum connman_device_mode mode;
 	enum connman_device_policy policy;
-	gboolean powered;
-	gboolean carrier;
-	gboolean scanning;
-	char *path;
+	connman_bool_t powered;
+	connman_bool_t carrier;
+	connman_bool_t scanning;
+	connman_uint8_t priority;
+	char *name;
+	char *node;
 	char *interface;
+	unsigned int connections;
 
 	struct connman_device_driver *driver;
 	void *driver_data;
 
+	connman_bool_t registered;
+
 	GHashTable *networks;
 };
 
@@ -125,7 +130,7 @@ static enum connman_device_policy string2policy(const char *policy)
 		return CONNMAN_DEVICE_POLICY_UNKNOWN;
 }
 
-static int set_powered(struct connman_device *device, gboolean powered)
+static int set_powered(struct connman_device *device, connman_bool_t powered)
 {
 	struct connman_device_driver *driver = device->driver;
 	int err;
@@ -239,6 +244,10 @@ static DBusMessage *get_properties(DBusConnection *conn,
 
 	DBG("conn %p", conn);
 
+	if (__connman_security_check_privilege(msg,
+					CONNMAN_SECURITY_PRIVILEGE_PUBLIC) < 0)
+		return __connman_error_permission_denied(msg);
+
 	reply = dbus_message_new_method_return(msg);
 	if (reply == NULL)
 		return NULL;
@@ -273,6 +282,10 @@ static DBusMessage *get_properties(DBusConnection *conn,
 		connman_dbus_dict_append_variant(&dict, "Policy",
 						DBUS_TYPE_STRING, &str);
 
+	if (device->priority > 0)
+		connman_dbus_dict_append_variant(&dict, "Priority",
+					DBUS_TYPE_BYTE, &device->priority);
+
 	connman_dbus_dict_append_variant(&dict, "Powered",
 					DBUS_TYPE_BOOLEAN, &device->powered);
 
@@ -314,11 +327,12 @@ static DBusMessage *set_property(DBusConnection *conn,
 	dbus_message_iter_next(&iter);
 	dbus_message_iter_recurse(&iter, &value);
 
-	if (__connman_security_check_privileges(msg) < 0)
+	if (__connman_security_check_privilege(msg,
+					CONNMAN_SECURITY_PRIVILEGE_MODIFY) < 0)
 		return __connman_error_permission_denied(msg);
 
 	if (g_str_equal(name, "Powered") == TRUE) {
-		gboolean powered;
+		connman_bool_t powered;
 		int err;
 
 		dbus_message_iter_get_basic(&value, &powered);
@@ -342,9 +356,15 @@ static DBusMessage *set_property(DBusConnection *conn,
 		err = set_policy(conn, device, policy);
 		if (err < 0)
 			return __connman_error_failed(msg);
+	} else if (g_str_equal(name, "Priority") == TRUE) {
+		connman_uint8_t priority;
+
+		dbus_message_iter_get_basic(&value, &priority);
+
+		device->priority = priority;
 	}
 
-	__connman_element_store(&device->element);
+	__connman_storage_save_device(device);
 
 	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
 }
@@ -354,7 +374,8 @@ static DBusMessage *create_network(DBusConnection *conn,
 {
 	DBG("conn %p", conn);
 
-	if (__connman_security_check_privileges(msg) < 0)
+	if (__connman_security_check_privilege(msg,
+					CONNMAN_SECURITY_PRIVILEGE_MODIFY) < 0)
 		return __connman_error_permission_denied(msg);
 
 	return __connman_error_invalid_arguments(msg);
@@ -365,7 +386,8 @@ static DBusMessage *remove_network(DBusConnection *conn,
 {
 	DBG("conn %p", conn);
 
-	if (__connman_security_check_privileges(msg) < 0)
+	if (__connman_security_check_privilege(msg,
+					CONNMAN_SECURITY_PRIVILEGE_MODIFY) < 0)
 		return __connman_error_permission_denied(msg);
 
 	return __connman_error_invalid_arguments(msg);
@@ -391,6 +413,9 @@ static DBusMessage *propose_scan(DBusConnection *conn,
 	if (!device->driver || !device->driver->scan)
 		return __connman_error_not_supported(msg);
 
+	if (device->powered == FALSE)
+		return __connman_error_failed(msg);
+
 	err = device->driver->scan(device);
 	if (err < 0)
 		return __connman_error_failed(msg);
@@ -464,6 +489,8 @@ static int register_interface(struct connman_element *element)
 		return -EIO;
 	}
 
+	device->registered = TRUE;
+
 	emit_devices_signal();
 
 	return 0;
@@ -471,8 +498,12 @@ static int register_interface(struct connman_element *element)
 
 static void unregister_interface(struct connman_element *element)
 {
+	struct connman_device *device = element->device;
+
 	DBG("element %p name %s", element, element->name);
 
+	device->registered = FALSE;
+
 	emit_devices_signal();
 
 	g_dbus_unregister_interface(connection, element->path,
@@ -510,6 +541,80 @@ static void device_disable(struct connman_device *device)
 		device->driver->disable(device);
 }
 
+static int setup_device(struct connman_device *device)
+{
+	int err;
+
+	DBG("device %p", device);
+
+	err = register_interface(&device->element);
+	if (err < 0) {
+		if (device->driver->remove)
+			device->driver->remove(device);
+		device->driver = NULL;
+		return err;
+	}
+
+	device_enable(device);
+
+	return 0;
+}
+
+static void probe_driver(struct connman_element *element, gpointer user_data)
+{
+	struct connman_device_driver *driver = user_data;
+
+	DBG("element %p name %s", element, element->name);
+
+	if (element->device == NULL)
+		return;
+
+	if (element->device->driver != NULL)
+		return;
+
+	if (driver->probe(element->device) < 0)
+		return;
+
+	element->device->driver = driver;
+
+	setup_device(element->device);
+}
+
+static void remove_device(struct connman_device *device)
+{
+	DBG("device %p", device);
+
+	device_disable(device);
+
+	unregister_interface(&device->element);
+
+	if (device->driver->remove)
+		device->driver->remove(device);
+
+	device->driver = NULL;
+}
+
+static void remove_driver(struct connman_element *element, gpointer user_data)
+{
+	struct connman_device_driver *driver = user_data;
+
+	DBG("element %p name %s", element, element->name);
+
+	if (element->device == NULL)
+		return;
+
+	if (element->device->driver == driver)
+		remove_device(element->device);
+}
+
+connman_bool_t __connman_device_has_driver(struct connman_device *device)
+{
+	if (device == NULL || device->driver == NULL)
+		return FALSE;
+
+	return device->registered;
+}
+
 static GSList *driver_list = NULL;
 
 static gint compare_priority(gconstpointer a, gconstpointer b)
@@ -535,30 +640,12 @@ int connman_device_driver_register(struct connman_device_driver *driver)
 	driver_list = g_slist_insert_sorted(driver_list, driver,
 							compare_priority);
 
-	//__connman_driver_rescan(&device_driver);
+	__connman_element_foreach(NULL, CONNMAN_ELEMENT_TYPE_DEVICE,
+						probe_driver, driver);
 
 	return 0;
 }
 
-static void remove_driver(struct connman_element *element, gpointer user_data)
-{
-	struct connman_device_driver *driver = user_data;
-
-	DBG("element %p name %s", element, element->name);
-
-	if (element->device == NULL)
-		return;
-
-	if (element->device->driver == driver) {
-		device_disable(element->device);
-
-		if (driver->remove)
-			driver->remove(element->device);
-
-		element->device->driver = NULL;
-	}
-}
-
 /**
  * connman_device_driver_unregister:
  * @driver: device driver definition
@@ -592,7 +679,8 @@ static void device_destruct(struct connman_element *element)
 
 	DBG("element %p name %s", element, element->name);
 
-	g_free(device->path);
+	g_free(device->node);
+	g_free(device->name);
 	g_free(device->interface);
 
 	g_hash_table_destroy(device->networks);
@@ -612,6 +700,7 @@ struct connman_device *connman_device_create(const char *node,
 						enum connman_device_type type)
 {
 	struct connman_device *device;
+	const char *str;
 
 	DBG("node %s type %d", node, type);
 
@@ -627,30 +716,41 @@ struct connman_device *connman_device_create(const char *node,
 	device->element.type = CONNMAN_ELEMENT_TYPE_DEVICE;
 	device->element.index = -1;
 
-	switch (type) {
-	case CONNMAN_DEVICE_TYPE_UNKNOWN:
-	case CONNMAN_DEVICE_TYPE_VENDOR:
-	case CONNMAN_DEVICE_TYPE_WIFI:
-	case CONNMAN_DEVICE_TYPE_WIMAX:
-	case CONNMAN_DEVICE_TYPE_BLUETOOTH:
-	case CONNMAN_DEVICE_TYPE_HSO:
-	case CONNMAN_DEVICE_TYPE_NOZOMI:
-	case CONNMAN_DEVICE_TYPE_HUAWEI:
-	case CONNMAN_DEVICE_TYPE_NOVATEL:
-		device->element.subtype = CONNMAN_ELEMENT_SUBTYPE_UNKNOWN;
-		break;
-	case CONNMAN_DEVICE_TYPE_ETHERNET:
-		device->element.subtype = CONNMAN_ELEMENT_SUBTYPE_ETHERNET;
-		break;
-	}
-
 	device->element.device = device;
 	device->element.destruct = device_destruct;
 
+	str = type2string(type);
+	if (str != NULL)
+		connman_element_add_static_property(&device->element,
+					"Type", DBUS_TYPE_STRING, &str);
+
 	device->type   = type;
 	device->mode   = CONNMAN_DEVICE_MODE_UNKNOWN;
 	device->policy = CONNMAN_DEVICE_POLICY_AUTO;
 
+	switch (type) {
+	case CONNMAN_DEVICE_TYPE_UNKNOWN:
+	case CONNMAN_DEVICE_TYPE_VENDOR:
+		device->priority = 0;
+		break;
+	case CONNMAN_DEVICE_TYPE_ETHERNET:
+	case CONNMAN_DEVICE_TYPE_WIFI:
+		device->priority = 100;
+		break;
+	case CONNMAN_DEVICE_TYPE_WIMAX:
+		device->priority = 20;
+		break;
+	case CONNMAN_DEVICE_TYPE_BLUETOOTH:
+		device->priority = 50;
+		break;
+	case CONNMAN_DEVICE_TYPE_HSO:
+	case CONNMAN_DEVICE_TYPE_NOZOMI:
+	case CONNMAN_DEVICE_TYPE_HUAWEI:
+	case CONNMAN_DEVICE_TYPE_NOVATEL:
+		device->priority = 60;
+		break;
+	}
+
 	device->networks = g_hash_table_new_full(g_str_hash, g_str_equal,
 						g_free, unregister_network);
 
@@ -683,19 +783,14 @@ void connman_device_unref(struct connman_device *device)
 }
 
 /**
- * connman_device_set_path:
+ * connman_device_get_name:
  * @device: device structure
- * @path: path name
  *
- * Set path name of device
+ * Get unique name of device
  */
-void connman_device_set_path(struct connman_device *device, const char *path)
+const char *connman_device_get_name(struct connman_device *device)
 {
-	g_free(device->element.devpath);
-	device->element.devpath = g_strdup(path);
-
-	g_free(device->path);
-	device->path = g_strdup(path);
+	return device->element.name;
 }
 
 /**
@@ -706,7 +801,7 @@ void connman_device_set_path(struct connman_device *device, const char *path)
  */
 const char *connman_device_get_path(struct connman_device *device)
 {
-	return device->path;
+	return device->element.path;
 }
 
 /**
@@ -823,6 +918,15 @@ int connman_device_set_powered(struct connman_device *device,
 
 	g_dbus_send_message(connection, signal);
 
+	if (powered == FALSE)
+		return 0;
+
+	if (device->policy != CONNMAN_DEVICE_POLICY_AUTO)
+		return 0;
+
+	if (device->driver->scan)
+		device->driver->scan(device);
+
 	return 0;
 }
 
@@ -857,9 +961,8 @@ int connman_device_set_carrier(struct connman_device *device,
 
 		element = connman_element_create(NULL);
 		if (element != NULL) {
-			element->type    = CONNMAN_ELEMENT_TYPE_DHCP;
-			element->subtype = device->element.subtype;
-			element->index   = device->element.index;
+			element->type  = CONNMAN_ELEMENT_TYPE_DHCP;
+			element->index = device->element.index;
 
 			if (connman_element_register(element,
 							&device->element) < 0)
@@ -871,6 +974,62 @@ int connman_device_set_carrier(struct connman_device *device,
 	return 0;
 }
 
+static void connect_known_network(struct connman_device *device)
+{
+	struct connman_network *network = NULL;
+	GHashTableIter iter;
+	gpointer key, value;
+	unsigned int count = 0;
+
+	DBG("device %p", device);
+
+	g_hash_table_iter_init(&iter, device->networks);
+
+	while (g_hash_table_iter_next(&iter, &key, &value) == TRUE) {
+		connman_uint8_t old_priority, new_priority;
+		connman_uint8_t old_strength, new_strength;
+
+		count++;
+
+		if (connman_network_get_remember(value) == FALSE)
+			continue;
+
+		if (network == NULL) {
+			network = value;
+			continue;
+		}
+
+		old_priority = connman_network_get_uint8(network, "Priority");
+		new_priority = connman_network_get_uint8(value, "Priority");
+
+		if (new_priority != old_priority) {
+			if (new_priority > old_priority)
+				network = value;
+			continue;
+		}
+
+		old_strength = connman_network_get_uint8(network, "Strength");
+		new_strength = connman_network_get_uint8(value, "Strength");
+
+		if (new_strength > old_strength)
+			network = value;
+	}
+
+	if (network != NULL) {
+		int err;
+
+		err = connman_network_connect(network);
+		if (err == 0 || err == -EINPROGRESS)
+			return;
+	}
+
+	if (count > 0)
+		return;
+
+	if (device->driver && device->driver->scan)
+		device->driver->scan(device);
+}
+
 /**
  * connman_device_set_scanning:
  * @device: device structure
@@ -911,9 +1070,103 @@ int connman_device_set_scanning(struct connman_device *device,
 
 	g_dbus_send_message(connection, signal);
 
+	if (scanning == TRUE)
+		return 0;
+
+	if (device->connections > 0)
+		return 0;
+
+	if (device->policy != CONNMAN_DEVICE_POLICY_AUTO)
+		return 0;
+
+	connect_known_network(device);
+
 	return 0;
 }
 
+/**
+ * connman_device_set_string:
+ * @device: device structure
+ * @key: unique identifier
+ * @value: string value
+ *
+ * Set string value for specific key
+ */
+int connman_device_set_string(struct connman_device *device,
+					const char *key, const char *value)
+{
+	DBG("device %p key %s value %s", device, key, value);
+
+	if (g_str_equal(key, "Name") == TRUE) {
+		g_free(device->name);
+		device->name = g_strdup(value);
+	} else if (g_str_equal(key, "Node") == TRUE) {
+		g_free(device->node);
+		device->node = g_strdup(value);
+	}
+
+	return 0;
+}
+
+/**
+ * connman_device_get_string:
+ * @device: device structure
+ * @key: unique identifier
+ *
+ * Get string value for specific key
+ */
+const char *connman_device_get_string(struct connman_device *device,
+							const char *key)
+{
+	DBG("device %p key %s", device, key);
+
+	if (g_str_equal(key, "Name") == TRUE)
+		return device->name;
+	else if (g_str_equal(key, "Node") == TRUE)
+		return device->node;
+
+	return NULL;
+}
+
+static void set_flightmode(struct connman_element *element, gpointer user_data)
+{
+	struct connman_device *device = element->device;
+	connman_bool_t flightmode = GPOINTER_TO_UINT(user_data);
+	connman_bool_t powered;
+
+	DBG("element %p name %s", element, element->name);
+
+	if (device == NULL)
+		return;
+
+	powered = (flightmode == TRUE) ? FALSE : TRUE;
+
+	if (device->powered == powered)
+		return;
+
+	set_powered(device, powered);
+}
+
+int __connman_device_set_flightmode(connman_bool_t flightmode)
+{
+	DBG("flightmode %d", flightmode);
+
+	__connman_element_foreach(NULL, CONNMAN_ELEMENT_TYPE_DEVICE,
+				set_flightmode, GUINT_TO_POINTER(flightmode));
+
+	return 0;
+}
+
+void __connman_device_increase_connections(struct connman_device *device)
+{
+	device->connections++;
+}
+
+void __connman_device_decrease_connections(struct connman_device *device)
+{
+	device->connections--;
+}
+
 /**
  * connman_device_add_network:
  * @device: device structure
@@ -940,6 +1193,8 @@ int connman_device_add_network(struct connman_device *device,
 
 	__connman_network_set_device(network, device);
 
+	__connman_storage_load_network(network);
+
 	err = connman_element_register((struct connman_element *) network,
 							&device->element);
 	if (err < 0) {
@@ -993,6 +1248,8 @@ int connman_device_remove_network(struct connman_device *device,
  */
 int connman_device_register(struct connman_device *device)
 {
+	__connman_storage_load_device(device);
+
 	return connman_element_register(&device->element, NULL);
 }
 
@@ -1004,6 +1261,8 @@ int connman_device_register(struct connman_device *device)
  */
 void connman_device_unregister(struct connman_device *device)
 {
+	__connman_storage_save_device(device);
+
 	connman_element_unregister(&device->element);
 }
 
@@ -1044,13 +1303,15 @@ static int device_probe(struct connman_element *element)
 {
 	struct connman_device *device = element->device;
 	GSList *list;
-	int err;
 
 	DBG("element %p name %s", element, element->name);
 
 	if (device == NULL)
 		return -ENODEV;
 
+	if (device->driver != NULL)
+		return -EALREADY;
+
 	for (list = driver_list; list; list = list->next) {
 		struct connman_device_driver *driver = list->data;
 
@@ -1068,16 +1329,7 @@ static int device_probe(struct connman_element *element)
 	if (device->driver == NULL)
 		return -ENODEV;
 
-	err = register_interface(element);
-	if (err < 0) {
-		if (device->driver->remove)
-			device->driver->remove(device);
-		return err;
-	}
-
-	device_enable(device);
-
-	return 0;
+	return setup_device(device);
 }
 
 static void device_remove(struct connman_element *element)
@@ -1092,12 +1344,7 @@ static void device_remove(struct connman_element *element)
 	if (device->driver == NULL)
 		return;
 
-	device_disable(device);
-
-	unregister_interface(element);
-
-	if (device->driver->remove)
-		device->driver->remove(device);
+	remove_device(device);
 }
 
 static struct connman_driver device_driver = {
@@ -1108,12 +1355,120 @@ static struct connman_driver device_driver = {
 	.remove		= device_remove,
 };
 
+static int device_load(struct connman_device *device)
+{
+	GKeyFile *keyfile;
+	gchar *pathname, *data = NULL;
+	gsize length;
+	char *str;
+	int val;
+
+	DBG("device %p", device);
+
+	pathname = g_strdup_printf("%s/%s.conf", STORAGEDIR,
+							device->element.name);
+	if (pathname == NULL)
+		return -ENOMEM;
+
+	keyfile = g_key_file_new();
+
+	if (g_file_get_contents(pathname, &data, &length, NULL) == FALSE) {
+		g_free(pathname);
+		return -ENOENT;
+	}
+
+	g_free(pathname);
+
+	if (g_key_file_load_from_data(keyfile, data, length,
+							0, NULL) == FALSE) {
+		g_free(data);
+		return -EILSEQ;
+	}
+
+	g_free(data);
+
+	str = g_key_file_get_string(keyfile, "Configuration", "Policy", NULL);
+	if (str != NULL) {
+		device->policy = string2policy(str);
+		g_free(str);
+	}
+
+	val = g_key_file_get_integer(keyfile, "Configuration",
+							"Priority", NULL);
+	if (val > 0)
+		device->priority = val;
+
+	g_key_file_free(keyfile);
+
+	return 0;
+}
+
+static int device_save(struct connman_device *device)
+{
+	GKeyFile *keyfile;
+	gchar *pathname, *data = NULL;
+	gsize length;
+	const char *str;
+
+	DBG("device %p", device);
+
+	pathname = g_strdup_printf("%s/%s.conf", STORAGEDIR,
+							device->element.name);
+	if (pathname == NULL)
+		return -ENOMEM;
+
+	keyfile = g_key_file_new();
+
+	if (g_file_get_contents(pathname, &data, &length, NULL) == FALSE)
+		goto update;
+
+	if (length > 0) {
+		if (g_key_file_load_from_data(keyfile, data, length,
+							0, NULL) == FALSE)
+			goto done;
+	}
+
+	g_free(data);
+
+update:
+	str = policy2string(device->policy);
+	if (str != NULL)
+		g_key_file_set_string(keyfile, "Configuration", "Policy", str);
+
+	if (device->priority > 0)
+		g_key_file_set_integer(keyfile, "Configuration",
+						"Priority", device->priority);
+
+	data = g_key_file_to_data(keyfile, &length, NULL);
+
+	g_file_set_contents(pathname, data, length, NULL);
+
+done:
+	g_free(data);
+
+	g_key_file_free(keyfile);
+
+	g_free(pathname);
+
+	return 0;
+}
+
+static struct connman_storage device_storage = {
+	.name		= "device",
+	.priority	= CONNMAN_STORAGE_PRIORITY_LOW,
+	.device_load	= device_load,
+	.device_save	= device_save,
+};
+
 int __connman_device_init(void)
 {
 	DBG("");
 
 	connection = connman_dbus_get_connection();
 
+	if (connman_storage_register(&device_storage) < 0)
+		connman_error("Failed to register device storage");
+
 	return connman_driver_register(&device_driver);
 }
 
@@ -1123,5 +1478,7 @@ void __connman_device_cleanup(void)
 
 	connman_driver_unregister(&device_driver);
 
+	connman_storage_unregister(&device_storage);
+
 	dbus_connection_unref(connection);
 }
